"""game.py - a context for running a game session or level

This file is part of Gummworld2.

Gummworld2 is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Gummworld2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.
"""

import pygame
from pygame.locals import (
    Rect, Color, ACTIVEEVENT,
    K_ESCAPE, K_LEFT, K_RIGHT, K_UP, K_DOWN, K_SPACE,
    K_w, K_a, K_s, K_d,
)

import gummworld2
from gummworld2 import State, BasicMapRenderer, Vec2d, Statf
from gummworld2 import model, toolkit

import settings, terrain, terrainnodes, player, rules
from terrainnodes import Air, Dirt, Backing
from terrainnodes import Air, Dirt, Backing


NEIGHBORS = (
    (-1,-1),
    ( 0,-1),
    ( 1,-1),
    ( 1, 0),
    ( 1, 1),
    ( 0, 1),
    (-1, 1),
    (-1, 0),
)


class Game(gummworld2.Engine):
    
    def __init__(self):
        
        window_title = 'Gummcraft - Space,Up=Jump | Tab=Tool | Mouse1=Dig | Mouse2=Build | Shift=Sprint'
        
        State.speed = settings.player_speed
        self.player = player.Player((0,0))
        
        gummworld2.Engine.__init__(self,
            caption=window_title,
            resolution=settings.resolution,
            camera_target=self.player,
            map=State.map,
            update_speed=settings.update_speed,
            frame_speed=settings.frame_speed,
        )
        State.clock.paused_callback = self.paused
        self.player.rect.centerx = settings.resolution[0]/2
        self.player.rect.bottom = State.map.SEA_LEVEL - 200
        self.player.position = self.player.rect.center
        State.camera.init_position(self.player.position)
        
        State.renderer = self.renderer = BasicMapRenderer(self.map)
        
        self.movex = 0
        self.sprint = 0
        self.mouse_down_1 = None
        self.mouse_down_3 = None
        
        toolkit.make_hud()
        next_pos = State.hud.next_pos
        State.hud.add('playerpos', Statf(next_pos(),
            'Player %s', callback=lambda:str(self.player.rect.center),
            interval=0.2))
        State.hud.add('tool', Statf(next_pos(),
            'Using %s', callback=lambda:str(self.player.tool.name),
            interval=0.2))
        
        State.game = self
        
        ## Lighting state. (Note: There's new lighting code also in
        ## terrainnodes/_node.py.)
        
        node_size = Vec2d(terrainnodes.NODE_SIZE())
        self.lighting_rect = pygame.Rect((0,0), node_size*3)
        self.visible = []
        self.last_screen_rect = pygame.Rect(self.camera.rect)
        self.dirty_rects = []
        self.renderer.set_rect(center=State.camera.rect.center)
        self.update_lighting(full_screen=True)
        
        # State for mouse focus helper.
        self.do_focus = False
    
    def update(self, dt):
        if pygame.KMOD_SHIFT & pygame.key.get_mods():
            self.sprint = 2 # run = speed * 2
        else:
            self.sprint = 1 # walk = speed * 1
        
        ## Calculate visible objects that may have changed due to mouse clicks
        ## or camera movement in order to apply lighting.
        dirty = rules.update(dt)
        State.camera.update()
        self.update_lighting(add_tiles=dirty)
        
        State.renderer.set_rect(center=State.camera.rect.center)
        State.hud.update(dt)
    
    def update_lighting(self, full_screen=False, add_tiles=[]):
        """update lighting on uncovered tiles
        - The default set is those tiles that are uncovered by panning the camera.
        - The value of full_screen is passed through to visible_generator().
        - The tile objects in add_tiles are added to the default set of tiles.
        """
        layer = State.map.layers[0]
        get_objects = layer.objects.intersect_objects
        set_dirty = self.renderer.set_dirty
        light_rect = self.lighting_rect
        dirty_rects = self.dirty_rects
        del dirty_rects[:]
        del self.visible[:]
        
        # Query additional tiles (e.g. those from mouse clicks).
        for obj in add_tiles:
            light_rect.center = obj.rect.center
            self.visible.extend(get_objects(light_rect))
        
        # Query default set of tiles.
        self.visible.extend(self.visible_generator(full_screen))
        
        # Loop tiles. Basically, any solid tile touching an air tile will be
        # marked light; otherwise it will be marked dark.
        for tile in self.visible:
            light_rect.center = tile.rect.center
            if terrainnodes.issolid(tile):
                neighbors = get_objects(light_rect)
                light = 0.6
                for n in neighbors:
                    if isinstance(n, Air):
                        light = 1.0
                        break
                if tile.set_lighting(light):
                    dirty_rects.append(tile.rect)
        
        # Update the renderer.
        set_dirty(*dirty_rects)
    
    def visible_generator(self, full_screen=True):
        """yield objects in newly revealed areas of the screen"""
        get_objects = State.map.layers[0].objects.intersect_objects
        if full_screen:
            visible_rects = [v.rect for v in self.renderer._visible_tiles]
        else:
            lsr = self.last_screen_rect
            sr = self.camera.rect
            visible_rects = []
            # Calculate a rect for X-axis movement.
            xmoved = False
            x,w = sr.x,sr.w
            if sr.x > lsr.x:
                x,w = sr.right,sr.x-lsr.x
                visible_rects.append(Rect(x,lsr.y,w,sr.h))
                xmoved = True
            elif sr.x < lsr.x:
                x,w = sr.x,lsr.x-sr.x
                visible_rects.append(Rect(x,lsr.y,w,sr.h))
                xmoved = True
            # Calculate a rect for Y-axis movement.
            ymoved = False
            y,h = sr.y,sr.h
            if sr.y > lsr.y:
                y,h = sr.bottom,sr.y-lsr.y
                visible_rects.append(Rect(lsr.x,y,sr.w,h))
                ymoved = True
            elif sr.y < lsr.y:
                y,h = sr.y,lsr.y-sr.y
                visible_rects.append(Rect(lsr.x,y,sr.w,h))
                ymoved = True
            # If X and Y changed, calculate the corner.
            if xmoved and ymoved:
                visible_rects.insert(0, Rect(x,y,w,h))
        # Save the current camera info.
        self.last_screen_rect[:] = self.camera.rect[:]
        # Query the objects from the map.
        for rect in visible_rects:
            objects = get_objects(rect)
#            objects.sort(key=lambda x:(x.rect.y,x.rect.x))
            for o in objects:
                yield o
    
    def draw(self, interp):
        State.screen.clear()
        State.renderer.draw_tiles()
        self.draw_player()
        self.draw_lighting()
        State.hud.draw()
        State.screen.flip()
    
    def draw_player(self):
        # playerpos - cameratopleft = undo interpolation
        camera = State.camera
        surf = camera.surface
        player = self.player
        dx,dy = camera.steady_target_position
        surf.blit(player.image, player.rect.move(dx,dy))
        
        # debug: draw mining tool hitbox
        if player.tool.draw_hitbox:
            r = player.tool.rect
            px,py = Vec2d(r.center) - camera.rect.center
            pygame.draw.rect(surf, Color('white'), r.move(dx,dy), 1)
    
    def draw_lighting(self):
        # debug: draw dirty rects for lighting
        if settings.debug_lighting:
            for r in self.dirty_rects:
                px,py = Vec2d(r.center) - camera.rect.center
                pygame.draw.rect(surf, Color('white'), r.move(dx,dy), 1)
    
    def on_key_down(self, unicode, key, mod):
        keys = settings.KEYS
        if key in keys['left']:    self.movex += -State.speed
        elif key in keys['right']: self.movex +=  State.speed
        elif key in keys['jump']: self.player.jump()
        elif key in keys['swaptool']: self.player.next_tool()
        elif key == K_ESCAPE: self.stop()
        elif key == pygame.K_RETURN:
            for area in [v.rect for v in self.renderer._visible_tiles]:
                for o in State.map.layers[0].objects.intersect_objects(area):
                    if isinstance(o, Dirt):
                        print o.rect.topleft,o.lighting,o.image_idx
            print '----------'
    
    def on_key_up(self, key, mod):
        keys = settings.KEYS
        if key in keys['left']:    self.movex -= -State.speed
        elif key in keys['right']: self.movex -=  State.speed
    
    def on_mouse_button_down(self, pos, button):
        if button == 1:
            self.mouse_down_1 = pos
        elif button == 3:
            self.mouse_down_3 = pos
    
    def on_mouse_button_up(self, pos, button):
        if button == 1:
            self.mouse_down_1 = None
        elif button == 3:
            self.mouse_down_3 = None
    
    def on_mouse_motion(self, pos, rel, buttons):
        if buttons[0] and self.mouse_down_1 is not None:
            self.mouse_down_1 = pos
        elif buttons[2] and self.mouse_down_3 is not None:
            self.mouse_down_3 = pos
    
    def on_active_event(self, gain, state):
        if gain == 0:
            # Mouse exited pygame window; auto-pause
            if self.do_focus:
                State.clock.pause()
            else:
                # This works around some kind of weird behavior when the
                # window first opens and the mouse position is polled.
                # Sometimes pygame thinks the mouse has left the window
                # area, or started outside the window area or something,
                # firing an ACTIVEEVENT with gain=0. Then you must drag
                # the mouse out and back in to receive gain=1.
                self.do_focus = True
        else:
            # Mouse entered pygame window; auto-unpause
            State.clock.resume()
    
    def paused(self):
        e = pygame.event.wait()
        if e.type == ACTIVEEVENT:
            if e.gain == 1:
                # Mouse entered pygame window; reset some stuff
                # Need to check if pressed, and set the value appropriately.
                keys = pygame.key.get_pressed()
                self.movex = 0
                for i in settings.KEYS['left']:
                    if keys[i]:
                        self.movex -= State.speed
                for i in settings.KEYS['right']:
                    if keys[i]:
                        self.movex += State.speed
                self.mouse_down_1 = False
                self.mouse_down_3 = False
                State.clock.resume()
                pygame.event.clear()
    
    def on_quit(self):
        self.stop()
    
    def stop(self):
        self.pop()
