import pygame

from gummworld2 import Vec2d

import _types


class Node(object):
    """A node of substance in the world
    
    size is the size of the node in pixels.
    
    anchor is the attribute of the rect used when setting the "position"
    attribute.
    
    images is undefined. It must be created by the subclass as a list.
    """
    size = 16,16
    anchor = 'center'
    
    deformable = True
    max_health = 1000.0
    time_till_reset = 10.0
    
    images = None   # subclass: set this to a dict
    detail = None   # subclass: set this to 0.0 to N.0; original r,g,b * detail
    
    backing = None  # if None, this node gets ambient light
    
    def __init__(self, position, image_files=[], start_image=None, lighting=1.0):
        self.image = None
        self.rect = None
        self._position = Vec2d(position)
        self._start_image = start_image
        self.lighting = lighting
        
        if not self.images:
            self._load_images(image_files)
        
        if start_image is None:
            start_image = image_files[0]
        self.image_name = start_image
        self.image = self._get_image()
        self.rect = self.image.get_rect()
        self.position = position
        self.image_idx = 0
        
        self.health = self.max_health
        self.time_since_hit = 0.0
        
        self.dirty = False
    
    def _set_position(self, val):
        self._position[:] = val
        x = int(round(val[0]))
        y = int(round(val[1]))
        setattr(self.rect, self.anchor, (x,y))
    def _get_position(self):
        return self._position[:]
    position = property(_get_position, _set_position,
        "position is stored as a float vector. Setting position will also" +
        " update the pygame rect attribute according to the class or" +
        " instance attribute 'anchor'.")
    
    def _load_images(self, image_files):
        images = self.images
        for name in image_files:
            ## this would typically load images from files
            if name not in images:
                images[name] = {}
            im = pygame.Surface(self.size)
            im.fill(pygame.Color(name))
            self._apply_detail(im)
            for lighting in xrange(11):
                final = self._apply_lighting(im, lighting/10.0)
                self.images[name][lighting] = final
    
    def _apply_detail(self, image):
        color = image.get_at((0,0))
        for attr in ('r','g','b'):
            val = getattr(color, attr)
            val = int(round(val * self.detail))
            if val > 255:
                val = 255
            setattr(color, attr, val)
        w,h = image.get_size()
        pygame.draw.line(image, color, (w/3,h-1), (w,h-1), 1)
        pygame.draw.line(image, color, (w-1,h), (w-1,h/3), 1)
    
    def _apply_lighting(self, image, lighting):
        """set brightness of image based on value 0.0 to 1.0
        """
        dest = image.copy()
        n = 255 * lighting
        if n > 0:
            dest.fill((n,n,n), None, special_flags=pygame.BLEND_RGBA_SUB)
        return dest
    
    def lighten(self, n=0.1):
        """increment or decrement brightness
        
        n must be between 0.0 and 1.0, inclusive.
        """
        lighting = self.lighting
        lighting -= n
        if lighting > 1.0:
            lighting = 1.0
        elif lighting < 0.0:
            lighting = 0.0
        self.lighting = lighting
        self.image = self._get_image()
    
    def set_lighting(self, lighting):
        if lighting > 1.0:
            lighting = 1.0
        elif lighting < 0.0:
            lighting = 0.0
        if lighting != self.lighting:
            self.lighting = lighting
            self.image = self._get_image()
            return True
        else:
            return False
    
    def _get_image(self):
        n = 10 - int(round(self.lighting * 10.0))
        self.image_idx = n
        return self.images[self.image_name][n]
    
    def heal(self, dt):
        self.time_since_hit -= dt
        if self.time_since_hit <= 0.0:
            self.health = self.max_health
    
    def hit(self, damage):
        if not self.deformable:
            return
        
        # decrement health and refresh time to reset
        self.health -= damage
        self.time_since_hit = self.time_till_reset
        
        # defeated
        if self.health < 0.0:
            self.health = 0.0
            self.time_since_hit = 0.0
            return
